import sys
import os
import copy as cp
import xarray as xr
import panel as pn
import numpy as np

import holoviews as hv
from holoviews import streams, opts, dim
from holoviews.streams import Selection1D
from holoviews.streams import Stream,param

from time import sleep
from random import choice

import hvplot.xarray
import panel as pn
import pandas as pd
import bokeh, param
import hvplot.networkx as hvnx
import networkx as nx
from bokeh.models import HoverTool
from scipy.signal import medfilt
from scipy.integrate import simps
from scipy.integrate import quad
from scipy.interpolate import InterpolatedUnivariateSpline
from hyperspy.misc.physical_constants import R, a0, m0
from hyperspy.misc.elements import elements

from saving_routines import Saving_panel

from bokeh.palettes import Set3 

print('Importing - Elemental quantification tool for EELS spectra')
hv.extension('bokeh',logo = False)
try:
    root_css = [el for el in sys.path if r'whatEELS\Library\css' in el][0]
except Exception as e:
    print('No root for css found. Skipping loading')
    print(e)
else:
    css_file = '{}\\css_styling.css'.format(root_css)
    if css_file not in pn.config.css_files:
        pn.config.css_files.append(css_file)


def change_extension():
    sleep(5)
    hv.extension('bokeh')
'''
def hooks_plotly_surfaces1(plot,element):
        plot.handles['layout']['scene'] =\
        dict(xaxis = dict(backgroundcolor="lightgray",gridcolor="white",\
            showbackground=True,zerolinecolor="red",zerolinewidth = 1,\
            tickfont={'color':'indianred'},spikecolor ='indianred'),\
        yaxis = dict(backgroundcolor='whitesmoke',gridcolor="white",\
            showbackground=True,zerolinecolor="blue",zerolinewidth = 1,\
            tickfont={'color':'slateblue'},
            spikecolor = 'slateblue'),\
        zaxis = dict(backgroundcolor='darkgrey',gridcolor="white",\
            showbackground=True,zerolinecolor="green",zerolinewidth = 1,\
            tickfont={'color':'limegreen'},spikecolor ='limegreen'),\
        xaxis_title='\u03b5 /R',\
        yaxis_title = 'log(q \u22c5 a\u2080)\u00b2',\
        zaxis_title = 'd\u00b2f / (dEdq)',\
        aspectmode = 'manual',aspectratio = {'x':1,'y':1,'z':0.66})
        plot.handles['layout']['margin'] = dict(r=0, l=50, b=30, t=0)
        plot.handles['layout']['scene_camera'] = dict(eye=dict(x=1.3, y=1.3, z=1), center=dict(x=0, y=0, z=-0.1))
        #plot.handles['fig']['layout']['zaxis'] = {'hoverlabel':'right'}
        print(plot.handles['layout']['title'])
        
        
def hooks_plotly_surfacesf(plot,element):
        plot.handles['layout']['scene'] =\
        dict(xaxis = dict(backgroundcolor="lightgray",gridcolor="white",\
            showbackground=True,zerolinecolor="red",zerolinewidth = 1,\
            tickfont={'color':'indianred'},spikecolor ='indianred'),\
        yaxis = dict(backgroundcolor='whitesmoke',gridcolor="white",\
            showbackground=True,zerolinecolor="blue",zerolinewidth = 1,\
            tickfont={'color':'slateblue'},
            spikecolor = 'slateblue'),\
        zaxis = dict(backgroundcolor='darkgrey',gridcolor="white",\
            showbackground=True,zerolinecolor="green",zerolinewidth = 1,\
            tickfont={'color':'limegreen'},spikecolor ='limegreen'),\
        xaxis_title='\u03b5 /R',\
        yaxis_title = 'log(q \u22c5 a\u2080)\u00b2',\
        zaxis_title = 'F-factor [a.u.]',\
        aspectmode = 'manual',aspectratio = {'x':1,'y':1,'z':0.66})
        plot.handles['layout']['margin'] = dict(r=0, l=50, b=30, t=0)
        plot.handles['layout']['scene_camera'] = dict(eye=dict(x=1.3, y=1.3, z=1), center=dict(x=0, y=0, z=-0.1))
        #plot.handles['fig']['layout']['zaxis'] = {'hoverlabel':'right'}
        print(plot.handles['layout']['title'])
'''
def formatter(value):
    #Method to format the yaxis format (Electron Counts)
    if value >= 0:
        return '{: .1e}'.format(value)
    else:
        return '{:.1e}'.format(value)

def formatter4(value):
    return '{:9.2f}'.format(value)

class Quantification_app(param.Parameterized):
    elems_quant = param.ListSelector()
    elems_width = param.ObjectSelector('EmptyEdgeSelection',objects=['EmptyEdgeSelection'])
    elems_extra_corr = param.ObjectSelector('NoneAvailable',objects=['NoneAvailable'])
    subshells_for_correction = param.ListSelector(default = [],objects=[])
    #subshell = param.ObjectSelector()
    #Surfaces to be used in the integral x_section calculation
    xsec = param.ObjectSelector('F-factor',\
    objects=['theoretical','beta-cut','F-factor'])
    corrections = param.Boolean(default=False)
    data_int = param.ObjectSelector('Raw - data',objects=['Raw - data','Fit - data'])
    int_window_ev = param.Number(default=0,step = 0.01) 
    onset_window = param.Number(default=0,step = 0.01)
    end_window = param.Number(default=0,step = 0.01)
    quantification_mode = param.ObjectSelector('Auto',objects=['Auto','Manual'])
    width_config_mode = param.ObjectSelector('Fixed',objects = ['Fixed','Fluid'])
    edges_for_quant = param.ListSelector([],objects=[])
    show_widths = param.Boolean(default=False)
    dictionary_for_int = param.Dict(default={})
    
    def __init__(self,ds,ccs_func,elems,factors):
        '''When launching this app, we should already have calculated the surfaces
        So let\'s use that info to get the surfaces
        
        The ds is going to be ds_first or ds_modi from the results panel
        '''
        super().__init__()
        self.dictio_windows = dict()
        for el in elems:
            self.dictio_windows[el] = list()
        self.ds = ds
        xsize = self.ds.x.values.size
        ysize = self.ds.y.values.size
        dist = abs(ysize-xsize)
        if xsize < ysize:
            self.xlims = (-0.5-dist/2,xsize-0.5+dist/2)
            self.ylims = (-0.5,ysize-0.5)
        else:
            self.ylims = (-0.5-dist/2,ysize-0.5+dist/2)
            self.xlims = (-0.5,xsize-0.5)
        self.get_initial_values(elems)
        '''
        lista_inicial = ['_'.join([ele,name]) for ele in self.elems_quant \
            for name in self.dictio_subshells_names[ele].keys()]
        self.param['edges_for_quant'].objects = self.dictio_subshells_names
        '''
        #self.dataset = ds_dataset
        self.x_sec_funcs = ccs_func
        self.scaling_factors = factors
        self.button_quantify = pn.widgets.Button(name = 'Quantify',width = 370,\
            margin = (5,15),button_type = 'success',disabled = True)
        self.button_quantify.on_click(self.quantify_elements)
        self.param['elems_quant'].objects = elems
        self.elems_quant = elems #A priori, all the elements selected for quantification
        '''
        self.param['elems_extra_corr'].objects = elems
        self.elems_extra_corr = elems[0]
        '''
        '''
        self.param['elems_width'].objects = elems
        self.elems_width = elems[0]
        '''
        #self.get_keywords()
        self.elementos_quant_wid = pn.Param(self.param['elems_quant'],\
            widgets = {'elems_quant':pn.widgets.CheckButtonGroup},\
            parameters = ['elems_quant'],show_labels = False,show_name = False,\
            width = 400)
        self.elementos_quant_wid[0].width = 370
        self.elementos_quant_wid[0].button_type = 'warning'
        self.elementos_gos_wid = pn.Param(self.param['elems_width'],\
            widgets = {'elems_width':pn.widgets.RadioButtonGroup},\
            parameters = ['elems_width'],show_labels = False,show_name = False,\
            width = 400)
        self.elementos_gos_wid[0].width = 370
        self.elementos_gos_wid[0].button_type = 'default'
        self.elementos_gos_wid[0].disabled = True
        self.edges_correction_width = pn.Param(self.param['elems_extra_corr'],\
            widgets = {'elems_extra_corr':pn.widgets.RadioButtonGroup},\
            parameters = ['elems_extra_corr'],show_labels = False,show_name = False,\
            width = 400)
        self.edges_correction_width[0].disabled = True
        self.quantification_mode_wid = pn.Param(self.param['quantification_mode'],\
            widgets = {'quantification_mode':pn.widgets.RadioButtonGroup},\
            parameters = ['quantification_mode'],show_labels = False,show_name = False,\
            width = 200,margin = (5,0))
        self.quantification_mode_wid[0].width = 180
        self.quantification_mode_wid[0].button_type = 'primary'
        self.quantification_mode_wid[0].disabled = True
        self.width_mode_wid = pn.Param(self.param['width_config_mode'],\
            widgets = {'width_config_mode':pn.widgets.RadioButtonGroup},\
            parameters = ['width_config_mode'],show_labels = False,show_name = False,\
            width = 200,margin = (5,0))
        self.width_mode_wid[0].width = 180
        self.width_mode_wid[0].button_type = 'warning'
        self.width_mode_wid[0].disabled = True
        self.quantification_mode_wid[0].margin = (5,0,5,15)
        self.width_mode_wid[0].margin = (5,15,5,0)
        self.edges_quant_selector_wid = pn.Param(self.param['edges_for_quant'],\
            widgets = {'edges_for_quant':pn.widgets.MultiChoice},\
            parameters = ['edges_for_quant'],show_labels = False,show_name = True,\
            name = 'Edges for quantification',width = 400)
        self.edges_quant_selector_wid[0].style = {'color':'white'}
        self.edges_quant_selector_wid[1].width = 350
        self.edges_quant_corrector_wid = pn.Param(self.param['subshells_for_correction'],\
            widgets = {'subshells_for_correction':pn.widgets.MultiChoice},\
            parameters = ['subshells_for_correction'],show_labels = False,show_name = True,\
            name = 'Edges for extra-count-corrections',width = 325)
        self.edges_quant_corrector_wid[0].style = {'color':'white'}
        self.edges_quant_corrector_wid[1].width = 300
        self.edges_quant_corrector_wid[1].height = 75
        self.edges_quant_corrector_wid[1].disabled = True
        ##############################################################################
        self.message_display = pn.pane.Markdown('No messages',\
            style = {'color':'white','text-align':'justify'},\
            width = 370,height = 200,margin = (5,15))
        self.info_panel = pn.Column(\
        pn.pane.Markdown('#### Info messages',style = {'color':'white'},\
            margin = (5,15,-10,15)),\
        pn.layout.Divider(height = 5,margin = (0,15,5,15)),\
        self.message_display,\
        pn.layout.Divider(height = 5,margin = (0,15)),\
        height = 315,width = 450,background = 'black',margin = (0,10,0,150))
        self.xsec_sel_wid = pn.Param(self.param['xsec'],\
            widgets = {'xsec':pn.widgets.RadioButtonGroup},\
            parameters = ['xsec'],show_labels = False,show_name = False,\
            width = 400)
        self.xsec_sel_wid[0].width = 370

        self.data_sel_wid = pn.Param(self.param['data_int'],\
            widgets = {'data_int':pn.widgets.RadioButtonGroup},\
            parameters = ['data_int'],show_labels = False,show_name = False,\
            width = 400)
        self.data_sel_wid[0].width = 370
        self.corrections_wid = pn.Param(self.param['corrections'],\
            widgets = {'corrections':pn.widgets.Toggle},parameters = ['corrections'],\
            width = 325)
        self.corrections_wid[0].name = 'No extra-corrections'
        self.corrections_wid[0].button_type = 'danger'
        self.corrections_wid[0].disabled = True
        self.button_info_corr = pn.widgets.Button(name = '\u2139',width = 25,\
            button_type = 'danger',margin = (10,20))
        self.button_info_corr.on_click(self._diplay_info_correction)
        #We need a toogle button to show the widths curretly selected for display
        self.button_show_int_windows = pn.Param(self.param['show_widths'],\
            widgets = {'show_widths':pn.widgets.Toggle},parameters = ['show_widths'],\
            width = 125, show_name = False, show_labels = False,margin = 0)
        self.button_show_int_windows[0].button_type = 'danger'
        self.button_show_int_windows[0].name = 'Show windows'
        self.button_show_int_windows[0].margin = (5,10,0,10)
        self.button_show_int_windows[0].disabled = True
        self.button_refresh_data = pn.widgets.Button(name = 'Update',\
            width = 75,margin = (23,5,5,5),button_type = 'success')
        self.button_refresh_data.on_click(self._update_values)
        self.button_reset_data = pn.widgets.Button(name = 'Reset \u21BA',\
            width = 75,margin = (23,5,5,5),button_type = 'danger')
        self.button_reset_data.on_click(self._reset_values)
        self.button_reset_data.disabled = True
        self.button_add_corrections = pn.widgets.Button(name = '\u2716',\
            disabled = True,width = 50,margin = (40,5,5,5),height = 40,\
            button_type = 'danger')
        self.button_add_corrections.on_click(self._add_corrections)
        #Let's buuld the other possibilities for integration - the widgets needed
        #onset of energy window
        self.onset_int_wid = pn.Param(self.param['onset_window'],\
            widgets = {'onset_window':pn.widgets.Spinner},parameters = ['onset_window'],\
            show_labels = False,show_name = True,\
            name = 'window onset-Eloss [eV]',width = 200,margin = 0)
        self.int_width_wid = pn.Param(self.param['int_window_ev'],\
            widgets = {'int_window_ev':pn.widgets.Spinner},\
            parameters = ['int_window_ev'],show_labels = False,\
            show_name = True,name = 'window width [eV]',\
            width = 200,margin = 0)
        self.end_int_wid = pn.Param(self.param['end_window'],\
            widgets = {'end_window':pn.widgets.Spinner},\
            parameters = ['end_window'],show_labels = False,\
            show_name = True,name = 'window Eloss-limit [eV]',\
            width = 200,margin = 0)
        self.onset_int_wid[1].disabled = True
        self.int_width_wid[1].disabled = True
        self.end_int_wid[1].disabled = True
        self.onset_int_wid[0].style = {'color':'grey'}
        self.int_width_wid[0].style = {'color':'grey'}
        self.end_int_wid[0].style = {'color':'grey'}
        self.onset_int_wid[0].margin = (0,15)
        self.int_width_wid[0].margin = (0,15)
        self.end_int_wid[0].margin = (0,15)
        self.label_int_width =\
        pn.pane.Markdown('window<br>width [eV]',width = 70,\
            style = {'color':'grey'},margin = (5,0))
        self.label_int_onset =\
        pn.pane.Markdown('window<br>onset [eV]',width = 70,\
            style = {'color':'grey'},margin = (5,0,5,5))
        #Let's create the image of the SI, as always, and the dynamic map
        TT_coords = [("x","@x"),("y","@y")]
        self.hover_tip_coords = HoverTool(tooltips=TT_coords)
        self.im = hv.Image(self.ds.OriginalData.sum('Eloss'))\
            .opts(cmap = 'greys_r',invert_yaxis = True,aspect = 'equal', frame_height=250,\
            xlim = self.xlims,ylim = self.ylims,\
            xaxis = None,yaxis = None,alpha = 0,shared_axes = False)
        self.hmap = hv.HeatMap(self.ds.OriginalData.sum('Eloss'))\
            .opts(cmap = 'greys_r',invert_yaxis = True,aspect = 'equal',frame_height=250,\
            xlim = self.xlims,ylim = self.ylims,\
            xaxis = None,yaxis = None,selection_fill_alpha = 1,\
            nonselection_fill_alpha = 0.75,line_width = 1.25,\
            line_color = 'grey',selection_line_color = 'red',\
            nonselection_line_color = 'grey',alpha = 0.75,\
            hover_fill_alpha = 1, hover_line_color = 'limegreen',\
            tools = ['tap',self.hover_tip_coords],shared_axes = False)
        self.hov_lims = (self.ds.x.values[0]-0.5,self.ds.x.values[-1]+\
                0.5,self.ds.y.values[0]-0.5,self.ds.y.values[-1]+0.5)
        empty_Data =\
            xr.Dataset({'EmptyData':(['y','x','Eloss'],\
            np.zeros_like(self.ds.OriginalData.values))},\
            coords = {'y':self.ds.y.values,'x':self.ds.x.values,\
            'Eloss':self.ds.Eloss.values})
        self.curve_empty = hv.Curve(empty_Data.EmptyData.isel(x = 0 ,y = 0))\
        .opts(frame_height = 250,frame_width = 650,show_grid = True,\
            shared_axes = False,yformatter=formatter,framewise = True,\
            line_width = 0.1,line_color = 'black')
        self.area_empty = hv.Area(empty_Data.EmptyData.isel(x = 0 ,y = 0))\
        .opts(frame_height = 250,frame_width = 650,show_grid = True,\
            shared_axes = False,yformatter=formatter,\
            line_width = 0.1,fill_alpha = 0.1,fill_color = 'grey',\
            line_color = 'black',framewise = True)
        self.create_dynamic_maps()
        
    def create_dynamic_maps(self):
        self.stream_tapping  = streams.SingleTap(x = -1,y = -1,source = self.im)
        self.dyn_data = hv.DynamicMap(self._plot_area_data,streams=[self.stream_tapping])\
            .opts(framewise = True,shared_axes = False,frame_height=250,frame_width = 650)
        self.dyn_fit_data = hv.DynamicMap(self._plot_fit_data,streams=[self.stream_tapping])\
            .opts(framewise = True,shared_axes = False,frame_height=250,frame_width = 650)
        self.dyn_raw_data = hv.DynamicMap(self._plot_raw_data,streams=[self.stream_tapping])\
            .opts(framewise = True,shared_axes = False,frame_height=250,frame_width = 650)
        
    def get_initial_values(self,elements_provided):
        dictio_fwhm_mult = dict()
        dictio_centers = dict()
        self.dictio_subshells_names = dict()
        dictio_E_int_limits = dict()
        self.dictio_E_int_lims_final = dict()
        for ele in elements_provided:
            #for ds_key in synthetic_list:
            dictio_fwhm_mult[ele] = dict()
            dictio_centers[ele] = dict()
            self.dictio_subshells_names[ele] = dict()
            dictio_E_int_limits[ele] = dict()
            for ds_key in self.ds:#app.res.ds_first:
                if all([word in ds_key for word in [ele,'cont','component']]):
                    if ds_key.split(ele)[1][0].islower():
                        '''#Do nothing- not intersting letter
                        This avoids problems with cases like having
                        Ce and C as elements of the same sample - the inial
                        letter may be followed by a lower case - problems then'''
                        pass
                    else:
                        sshell = ds_key.split('_')[0].split(ele)[1]
                        #lista_edges.append(sshell)
                        #They all follow the same order....
                        dictio_fwhm_mult[ele][sshell] = list()
                        dictio_centers[ele][sshell] = list()
                        self.dictio_subshells_names[ele][sshell] = list()
                        #print(ele,ds_key,sshell)
                        if len(sshell) == 3 and 'K' not in sshell:
                            #Now we have to look for the subshell components and assest the 
                            #values we are looking for
                            ssh1 = ''.join([ele,sshell[0],sshell[1]])
                            ssh2 = ''.join([ele,sshell[0],sshell[-1]])
                            #for possible in synthetic_list:
                            #NOw we have to look for those possible ELNES subshells in the components fitted
                            #otherwise - a continuum-like compo will be added as reference
                            for ssh  in [ssh1,ssh2]:
                                if '_'.join([ssh,'component']) in self.ds: #app.res.ds_first:
                                    fwhm_key   = '_'.join([ssh,'fwhm'])
                                    center_key = '_'.join([ssh,'center'])
                                    self.dictio_subshells_names[ele][sshell].append(ssh.split(ele)[1])
                                    dictio_fwhm_mult[ele][sshell].append(np.nanmedian(self.ds[fwhm_key].values))
                                    dictio_centers[ele][sshell].append(np.nanmedian(self.ds[center_key].values))
                            if all(['_'.join([ssh,'component']) not in self.ds for ssh in [ssh1,ssh2]]):
                                #The case of having actually removed all the sshell ELNES previously
                                #Just an standard apporach .... 10 times the channel dispersion and
                                #The theoretical onset
                                '''
                                dictio_fwhm_mult[ele][sshell] =\
                                    (app.res.ds_first.Eloss.values[1]-app.res.ds_first.Eloss.values[0])*10
                                dictio_centers[ele][sshell] = min([elements[ele]['Atomic_properties']\
                                    ['Binding_energies'][ssh]['onset_energy (eV)'] \
                                    for ssh in [ssh1,ssh2]])
                                '''
                                self.dictio_subshells_names[ele][sshell].append(sshell)
                                dictio_fwhm_mult[ele][sshell].append(\
                                    (self.ds.Eloss.values[1]-self.ds.Eloss.values[0])*10)
                                #The minumum of the onsets associated to this edge
                                # The elements dictionary must be imported from hyperspy
                                dictio_centers[ele][sshell].append(min([elements[ele]['Atomic_properties']\
                                    ['Binding_energies'][ssh]['onset_energy (eV)'] \
                                    for ssh in [ssh1.split(ele)[1],ssh2.split(ele)[1]]]))  
                        elif len(sshell) == 2:
                            #The same modus opperandi - look for valid subshells fitted as ELNES .... 
                            #if not, go to the reference values
                            if ''.join([ele,sshell,'_component']) in self.ds:#app.res.ds_first:
                                #In this case...there is only one possible ELNES feature...1 possible outcome
                                fwhm_key   = ''.join([ele,sshell,'_fwhm'])
                                center_key = ''.join([ele,sshell,'_center'])
                                self.dictio_subshells_names[ele][sshell].append(sshell)
                                dictio_fwhm_mult[ele][sshell].append(np.nanmedian(self.ds[fwhm_key].values))
                                dictio_centers[ele][sshell].append(np.nanmedian(self.ds[center_key].values))
                            else:
                                self.dictio_subshells_names[ele][sshell].append(sshell) #So we now no ELNES was in place
                                dictio_fwhm_mult[ele][sshell].append((self.ds.Eloss.values[1]-self.ds.Eloss.values[0])*10)
                                dictio_centers[ele][sshell].append(elements[ele]['Atomic_properties']\
                                    ['Binding_energies'][sshell]['onset_energy (eV)'])
                        elif 'K' in sshell:
                            '''This is a triky one..... always, always we will face 
                            a priori a single continuum component. If ELNES is added...then we must 
                            proceed with caution.
                            We are going to search all the K ELNES (artificially added in anay case)
                            So we can assert if we have a lower than onset energy for an ELNES structure,
                            and an stimated sigma value'''
                            marker = 0
                            for rest_ds in self.ds:#app.res.ds_first:
                                #Actually, let's put together element and sshell
                                cadena = ''.join([ele,'K'])
                                if all([word in rest_ds for word in [cadena,'component']])\
                                and 'cont' not in rest_ds:
                                    clave = rest_ds.split('_')[0]
                                    self.dictio_subshells_names[ele][sshell].append(clave)
                                    dictio_fwhm_mult[ele][sshell].append(np.nanmedian(self.ds['_'.join([clave,'fwhm'])].values))
                                    dictio_centers[ele][sshell].append(np.nanmedian(self.ds['_'.join([clave,'center'])].values))
                                    marker+=1
                            #Otherwise....if only the large
                            #continuum is in the model...we go with onset and 10 times res, as before
                            if marker == 0:
                                self.dictio_subshells_names[ele][sshell].append(sshell)
                                dictio_centers[ele][sshell].append(elements[ele]['Atomic_properties']\
                                    ['Binding_energies'][sshell]['onset_energy (eV)'])
                                dictio_fwhm_mult[ele][sshell].append((self.ds.Eloss.values[1]-self.ds.Eloss.values[0])*10)
                    #We need the values for each subshell added
                    val_fwhm = max(dictio_fwhm_mult[ele][sshell])
                    val_min = min(dictio_centers[ele][sshell])
                    val_max = max(dictio_centers[ele][sshell])
                    #We are setting the min-Evalue as the minimum between the spectra min E and the calculated val
                    #The same goes for the max-Evalue, with the spectrum limit
                    lim_inf = val_min - val_fwhm * 0.5 #half the val_fwhm calculated
                    lim_sup = val_max + val_fwhm*10
                    dictio_E_int_limits[ele][sshell] = [max(lim_inf,self.ds.Eloss.values[0]),\
                        min(lim_sup,self.ds.Eloss.values[-1])]
        #At the end of all...it is time to loop again and select the default settings for each subshell
        #Those would be the smaller windows, of the same size, for all the edges involved
        window_val = min([dictio_E_int_limits[ele][sshell][-1]-dictio_E_int_limits[ele][sshell][0] \
            for ele in elements_provided for sshell in dictio_E_int_limits[ele].keys()])
        for ele in elements_provided:
            self.dictio_E_int_lims_final[ele] = dict()
            for sshell in dictio_E_int_limits[ele].keys():
                self.dictio_E_int_lims_final[ele][sshell] =\
                [round(dictio_E_int_limits[ele][sshell][0],2),\
                round(dictio_E_int_limits[ele][sshell][0]+window_val,2)]
        #We are going to copy all these exact values in the other possible dictionaries - 
        #Case of manual width_adjustment, single value all along
        self.dictio_E_lims_manual_h = cp.deepcopy(self.dictio_E_int_lims_final)
        #Case of manual width adjustment, individual values for each element
        self.dictio_E_lims_manual_i = cp.deepcopy(self.dictio_E_int_lims_final)

    def _plot_fit_data(self,x,y):
        xi,xf,yi,yf = self.hov_lims
        if x >= xf or x <= xi:
            return self.curve_empty#.opts(line_alpha = 0)
        elif y >= yf or y <= yi:
            return self.curve_empty#.opts(line_alpha = 0)
        else:
            x0,y0 = self.im.closest((x,y))
            self.x_pos = int(x0)
            self.y_pos = int(y0)
            curva = hv.Curve(self.ds.BestFit.isel(x = int(x0),y = int(y0)))\
                .opts(show_grid=True,show_title=False,line_width=1.5,\
                line_color = 'limegreen',yformatter=formatter,
                shared_axes = False,framewise = True)
            return curva
        
    def _plot_raw_data(self,x,y):
        xi,xf,yi,yf = self.hov_lims
        if x >= xf or x <= xi:
            return self.curve_empty#.opts(line_alpha = 0)
        elif y >= yf or y <= yi:
            return self.curve_empty#.opts(line_alpha = 0)
        else:
            x0,y0 = self.im.closest((x,y))
            self.x_pos = int(x0)
            self.y_pos = int(y0)
            curva = hv.Curve(self.ds.OriginalData.isel(x = int(x0),y = int(y0)))\
                .opts(show_grid=True,show_title=False,line_width=1,\
                line_color = 'darkred',yformatter=formatter,
                shared_axes = False,framewise = True)
            return curva
    
    def _plot_area_data(self,x,y):
        xi,xf,yi,yf = self.hov_lims
        if x >= xf or x <= xi:
            return self.area_empty
        elif y >= yf or y <= yi:
            return self.area_empty
        else:
            x0,y0 = self.im.closest((x,y))
            self.x_pos = int(x0)
            self.y_pos = int(y0)
            area = hv.Area(self.ds.OriginalData.isel(x = int(x0),y = int(y0)))\
                .opts(show_grid=True,show_title=False,line_width=0.1,\
                yformatter=formatter,line_color = 'black',\
                shared_axes = False,framewise = True,fill_color = 'grey',fill_alpha = 0.5)
            return area
    
    
    def get_datasets_quantified(self):
        dictio_dataset = dict()
        xs = self.ds.x.values
        ys = self.ds.y.values
        for elem in self.elems_quant:
            clave1 = '_'.join([elem,'_integrated_intensity'])
            #clave2 = '_'.join([elem,'_xsection'])
            dictio_dataset[clave1] = (['y','x'],cp.deepcopy(self.dictio_integrated_intensity[elem]))
            #dictio_dataset[clave2] = (['y','x'],cp.deepcopy(self.dictio_xsec[elem]['val']))
        if self.corrections:
            for el in self.dictio_corrected_intensity:
                try:
                    clave3 = '_'.join([el,'_corrected_intensity'])
                    dictio_dataset[clave3] = (['y','x'],cp.deepcopy(self.dictio_corrected_intensity[el]))
                except:
                    pass
        else: pass
        #And now the creation of the dataset
        #We need a mask, to avoid showing datapoints in conflict with the fact of having adjusted stuff
        #So, for one of the elements in the quantification, let's check the continuum component, and in those
        #nan positions we place a 0
        clav1 = ''.join(self.edges_for_quant[0].split('_'))
        mask = np.nan_to_num(self.ds[''.join([clav1,'_cont_component'])].values,nan=-1).sum(-1)
        mask[mask == -1*self.ds.Eloss.size] = 0
        dictio_dataset['mask'] = (['y','x'],mask)
        self.integrations_dataset = xr.Dataset(dictio_dataset,coords={'x':xs,'y':ys})
            
    
    def quantify_elements(self,event):
        if not self.corrections:
            self.button_quantify.button_type = 'warning'
            self.button_quantify.name = 'Integrating - wait'
            self.button_quantify.disabled = True
            self.quantification_classic()
            self.button_quantify.name = 'Quantify'
            self.button_quantify.button_type = 'success'
            self.button_quantify.disabled = False
            self.get_datasets_quantified()
            #Let's get the dataset of the quantifications
            self.shRes = Show_results(self.ds,self.integrations_dataset,\
                self.dictio_xsec,self.edges_for_quant,self.data_int,\
                self.quantification_mode,self.width_config_mode)
            self.shRes.create_display()
            self.shRes.display.show()
        else:
            #By now, do nothing
            self.button_quantify.button_type = 'warning'
            self.button_quantify.name = 'Integrating - wait'
            self.button_quantify.disabled = True
            self.quantification_classic()
            self.calculate_correction_factors()
            self.calculate_corrected_counts()
            self.button_quantify.name = 'Quantify'
            self.button_quantify.button_type = 'success'
            self.button_quantify.disabled = False
            self.get_datasets_quantified()
            self.shRes = Show_results(self.ds,self.integrations_dataset,\
                self.dictio_xsec,self.edges_for_quant,self.data_int,\
                self.quantification_mode,self.width_config_mode,True,\
                self.dictionary_extra_corrections)
            self.shRes.create_display()
            self.shRes.display.show()
        
        #At the end.....whatever case we are in...we need to assembly the whole dataset
        
    def calculate_correction_factors(self):
        #When we enable extra corrections.....this happens -
        self.correction_factors = dict()
        for elem in self.dictionary_extra_corrections:
            self.correction_factors[elem] = dict()
            for sshell in self.dictionary_extra_corrections[elem]:
                ele,ssh = sshell.split('_')
                #NOw...ensure that we are using a valid reference area for the subshell comparison, otherwise do not use those corrections
                if not any(['_'.join([ele,ssh[0]]) in word for word in self.edges_for_quant]):
                    #If no valid reference in place in the integration phase....skip the correction phase
                    pass 
                else:
                    #We also may have at this point the xsection integrated....so let's get the integrated x-sec for the subshell in the corrections
                    #To get the coefficient needed
                    try: 
                        ref = self.dictio_xsec[ele]['val']
                    except:
                        coeff = 0
                    else:
                        #Let's integreta the Gd part with the width that we have in the quantification
                        func = lambda E: self.x_sec_funcs[self.xsec][ele][ssh](E)*self.scaling_factors[ele][ssh]
                        a,b = self.dictionary_for_int[elem]
                        coeff = quad(func,a,b,limit = 100)[0] / self.dictio_xsec[ele]['val']
                    finally:
                        self.correction_factors[elem][ssh] = coeff
                        self.correction_factors[elem]['ref'] = ele
    
    def calculate_corrected_counts(self):
        #With the calculated correction extra factors..we get now the corrected counts for the elements
        self.dictio_corrected_intensity = dict()
        for el in self.correction_factors.keys():
            #We may have the case of an empty dictionary, given a defective data input from the user
            self.dictio_corrected_intensity[el] = cp.deepcopy(self.dictio_integrated_intensity[el])
            #if self.correction_factors[el] != dict():
            for ssh in self.correction_factors[el]:
                if ssh != 'ref':
                    ref = self.correction_factors[el]['ref']
                    ratio = self.correction_factors[el][ssh]
                    print(ratio)
                    '''We also have to be aware that the correction should only be
                    applied to the continuum components.... given that 
                    only that factor is theoretically  related with the coefficient
                    '''
                    #Let's find out the continuum component in play
                    for edge in self.edges_for_quant:
                        eli,sshell = edge.split('_')
                        if ref == eli:
                            clave = ''.join([ref,sshell,'_cont_component'])
                            #Now we have to integrate the values of the continuum components
                            #We get the integration limits
                            Einf,Esup = list(np.searchsorted(self.ds.Eloss,\
                                self.dictionary_for_int[ref]))
                            #We integrate by simps
                            mat_ints = simps(np.nan_to_num(self.ds[clave].values)[:,:,Einf:Esup],\
                                    self.ds.Eloss.values[Einf:Esup])
                            #print(clave,Einf,Esup)
                            #print(mat_ints,self.correction_factors[el][ssh])
                            mat_value = mat_ints*ratio
                            print('passed')
                            self.dictio_corrected_intensity[el] -= mat_value
            #At the end of the day....we want the integrated intensities to be >= 0, so let's see to it
            self.dictio_corrected_intensity[el] = np.maximum(self.dictio_corrected_intensity[el],0)
            #else:
            #    pass
        #self.dictio_integrated_intensity[el]

    def quantification_classic(self):
        #The xsections are only integrated once
        #xsec_selected = self.xsec
        self.dictio_xsec = dict()
        for ele in self.dictionary_for_int.keys():
            self.dictio_xsec[ele] = {'lims':list(),'val':0}
            for edge in self.dictio_subshells_names[ele]:
                if '_'.join([ele,edge]) in self.edges_for_quant:
                    self.dictio_xsec[ele]['lims'] = cp.deepcopy(self.dictionary_for_int[ele])
                    a,b = self.dictionary_for_int[ele]
                    for sshell in self.dictio_subshells_names[ele][edge]:
                        function_to_int = lambda E: self.x_sec_funcs[self.xsec][ele][sshell](E)*self.scaling_factors[ele][sshell]
                        self.dictio_xsec[ele]['val']+=\
                        quad(function_to_int,a,b,limit = 100)[0]
        #print('xsec_calculated')
        #Let's get now the intregral values using the specified data
        dictio_ints = dict()
        dictio_substractions = dict()
        self.dictio_integrated_intensity = dict()
        if self.data_int == 'Raw - data':
            #Here we have a two step approach. We get the whole curve integration, and
            #Substract the rest of the elements from it
            for el in self.dictionary_for_int.keys():
                #We get directly the indices
                Einf,Esup = list(np.searchsorted(self.ds.Eloss,\
                    self.dictionary_for_int[el]))
                res = simps(self.ds.OriginalData.values[:,:,Einf:Esup],self.ds.Eloss.values[Einf:Esup])
                dictio_ints[el] = res
                rest = np.zeros(self.ds.OriginalData.values.shape[:-1])
                #print(Einf,Esup)
                for el2 in self.dictio_subshells_names:
                    for edge in self.dictio_subshells_names[el2]:
                        if '_'.join([el2,edge]) not in self.edges_for_quant or el2 != el:
                            compo = ''.join([el2,edge,'_cont_component'])
                            #print('Substracting {} to the raw data of {}'.format(compo,el))
                            rest += simps(np.nan_to_num(self.ds[compo].values)[:,:,Einf:Esup],\
                                self.ds.Eloss.values[Einf:Esup])
                            for sshell in self.dictio_subshells_names[el2][edge]:
                                if sshell != edge:
                                    compo = ''.join([el2,sshell,'_component'])
                                    #print('Substracting {} to the raw data of {}'.format(compo,el))
                                    rest += simps(np.nan_to_num(self.ds[compo].values)[:,:,Einf:Esup],\
                                        self.ds.Eloss.values[Einf:Esup])
                            #We have to get the whole continuum set onces, and the elnes structures
                dictio_substractions[el] = rest
                self.dictio_integrated_intensity[el] = np.maximum(res - rest,0) #Avoids negative counts - in those places it goes to 0 the integral
                
        elif self.data_int == 'Fit - data':
            #dictio_for_intregration_counts = dict()
            for edge in self.edges_for_quant:
                el,shell = edge.split('_')
                #First we get the continuum integration of counts
                clave_cont = ''.join([el,shell,'_cont_component'])
                E0,E1 = np.searchsorted(self.ds.Eloss.values,self.dictionary_for_int[el])
                integration_counts = simps(np.nan_to_num(self.ds[clave_cont].values)[:,:,E0:E1],self.ds.Eloss.values[E0:E1])
                #self.dictio_integrated_intensity[el] = conti_int
                for ssh in self.dictio_subshells_names[el][shell]:
                    try:
                        #This protects from errors when only a single continuum component is used
                        clav = ''.join([el,ssh,'_component'])
                        integration_counts += simps(np.nan_to_num(self.ds[clav].values)[:,:,E0:E1],self.ds.Eloss.values[E0:E1])
                    except:
                        pass
                self.dictio_integrated_intensity[el] = np.maximum(integration_counts,0)
            #print('Integrating Fit')
    '''
    def _get_data_structures_quantification(self):
        for el in self.elems_quant:
            pass
    ''' 
    
    def _diplay_info_correction(self,event):
        self.message_display.object = 'Allows to correct the raw electron counts\
        of the selected elements with non-used/fitted cross sections that may\
        have an influence on the elemental edges'
    
    def _changes_from_quantification_mode(self):
        if self.quantification_mode == 'Manual':
            self.elementos_gos_wid[0].disabled = False
            self.width_mode_wid[0].disabled = False
            self.onset_int_wid[1].disabled = False
            self.onset_int_wid[0].style = {'color':'limegreen'}
            self.button_refresh_data.disabled = False
            self.button_reset_data.disabled = False
            self._changes_from_width_config_mode()
        else: 
            #Case of automatic settings
            self.elementos_gos_wid[0].disabled = True
            self.width_mode_wid[0].disabled = True
            self.onset_int_wid[0].style = {'color':'grey'}
            self.int_width_wid[0].style = {'color':'grey'}
            self.end_int_wid[0].style = {'color':'grey'}
            self.int_width_wid[1].disabled = True
            self.onset_int_wid[1].disabled = True
            self.end_int_wid[1].disabled = True
            self._modify_current_dictio_for_int()
            
    def _changes_from_width_config_mode(self):
        el_curr,ssh_curr = self.elems_width.split('_')
        if self.width_config_mode == 'Fixed':
            self.int_width_wid[1].disabled = False
            self.end_int_wid[1].disabled = True
            self.end_int_wid[0].style = {'color':'grey'}
            self.int_width_wid[0].style = {'color':'limegreen'}
            self.int_window_ev =\
            round(self.dictio_E_lims_manual_h[el_curr][ssh_curr][1]\
            -self.dictio_E_lims_manual_h[el_curr][ssh_curr][0],2)
            self.onset_window = self.dictio_E_lims_manual_h[el_curr][ssh_curr][0]
            self._modify_current_dictio_for_int()
            
        elif self.width_config_mode == 'Fluid':
            self.int_width_wid[1].disabled = True
            self.end_int_wid[1].disabled = False
            self.end_int_wid[0].style = {'color':'limegreen'}
            self.int_width_wid[0].style = {'color':'grey'}
            self.end_window =\
            self.dictio_E_lims_manual_i[el_curr][ssh_curr][1]
            self.onset_window = self.dictio_E_lims_manual_i[el_curr][ssh_curr][0]
            self._modify_current_dictio_for_int()
    
    
    def _modify_current_dictio_for_int(self):
        if self.quantification_mode == 'Manual':
            if self.width_config_mode == 'Fixed':
                for ed in self.edges_for_quant:
                    el,sshell = ed.split('_')
                    self.dictionary_for_int[el] =\
                    cp.deepcopy(self.dictio_E_lims_manual_h[el][sshell])
            elif self.width_config_mode == 'Fluid':
                for ed in self.edges_for_quant:
                    el,sshell = ed.split('_')
                    self.dictionary_for_int[el] =\
                    cp.deepcopy(self.dictio_E_lims_manual_i[el][sshell])
        else:
            for ed in self.edges_for_quant:
                el,sshell = ed.split('_')
                self.dictionary_for_int[el] =\
                cp.deepcopy(self.dictio_E_int_lims_final[el][sshell])
            
    def _reset_values(self,event):
        #self.show_widths = False
        if self.quantification_mode == 'Manual':
            for ed in self.param['elems_width'].objects:
                el,sshell = ed.split('_')
                if self.width_config_mode == 'Fixed':
                    self.dictio_E_lims_manual_h[el][sshell] =\
                    self.dictio_E_int_lims_final[el][sshell]
                elif self.width_config_mode == 'Fluid':
                    self.dictio_E_lims_manual_i[el][sshell] =\
                    self.dictio_E_int_lims_final[el][sshell]
            #Finally we update values in the display
            el,sshell = self.elems_width.split('_')
            if self.width_config_mode == 'Fixed':
                self.onset_window = self.dictio_E_int_lims_final[el][sshell][0]
                self.int_window_ev = round(self.dictio_E_int_lims_final[el][sshell][1]-\
                self.dictio_E_int_lims_final[el][sshell][0],2)
            else:
                self.onset_window = self.dictio_E_int_lims_final[el][sshell][0]
                self.end_window = self.dictio_E_int_lims_final[el][sshell][1]
            self._modify_current_dictio_for_int()
        else:
            self._modify_current_dictio_for_int()
            pass
        self._add_integration_windows()
        
    def _update_values(self,event):
        if self.quantification_mode == 'Manual':
            if self.width_config_mode == 'Fixed':
                el,sshell = self.elems_width.split('_')
                self.dictio_E_lims_manual_h[el][sshell] =\
                [round(self.onset_window,2),\
                round(self.onset_window+self.int_window_ev,2)]
                for ed in self.param['elems_width'].objects:
                    if ed != self.elems_width:
                        ele,ssh = ed.split('_') 
                        self.dictio_E_lims_manual_h[ele][ssh] =  \
                        [self.dictio_E_lims_manual_h[ele][ssh][0],\
                        round(self.dictio_E_lims_manual_h[ele][ssh][0]\
                        +self.int_window_ev,2)]
            elif self.width_config_mode == 'Fluid':
                el,sshell = self.elems_width.split('_')
                self.dictio_E_lims_manual_i[el][sshell] =\
                [round(self.onset_window,2),round(self.end_window,2)]
            self._modify_current_dictio_for_int()
        else:
            self._modify_current_dictio_for_int()
        self._add_integration_windows()
    
    @param.depends('show_widths',watch = True)
    def _add_integration_windows(self):
        if len(list(self.dictionary_for_int.keys())) == 0:
            #No vertical int windows to be shown...so do nothing
            #Reset to only show the dynamic maps in any case
            if self.data_int == 'Raw - data':
                self.placeholder_dynamic.object = self.dyn_data*self.dyn_raw_data
            else:
                self.placeholder_dynamic.object = self.dyn_data*self.dyn_fit_data
            self.button_show_int_windows[0].button_type = 'danger'
            return
        if self.show_widths:
            #And we then create the span objects and append it to the display
            self.button_show_int_windows[0].button_type = 'success'
            lista_spans = []
            for el in self.dictionary_for_int.keys():
                x1 = self.dictionary_for_int[el][0]
                x2 = self.dictionary_for_int[el][1]
                lista_spans.append(hv.VSpan(x1 = x1,x2 = x2)\
                    .opts(shared_axes = False,frame_width=650,frame_height=250,\
                    fill_alpha = 0.5,fill_color = 'gold',line_color = 'orange',\
                    line_alpha = 1))
            self.int_span = lista_spans[0]
            if len(lista_spans)>1:
                for sp in lista_spans[1:]:
                    self.int_span*=sp
            #And we set the spans in the graph
            if self.data_int == 'Raw - data':
                self.placeholder_dynamic.object =\
                self.dyn_data*self.dyn_raw_data*self.int_span
            else:
                self.placeholder_dynamic.object =\
                self.dyn_data*self.dyn_fit_data*self.int_span
        else:
            self.button_show_int_windows[0].button_type = 'danger'
            if self.data_int == 'Raw - data':
                self.placeholder_dynamic.object = self.dyn_data*self.dyn_raw_data
            else:
                self.placeholder_dynamic.object = self.dyn_data*self.dyn_fit_data
                
    @param.depends('data_int',watch = True)
    def _change_data_selected(self):
        ['Raw - data','Fit - data']
        if self.data_int == 'Raw - data':
            if self.show_widths:
                self._add_integration_windows()
            else:
                self.placeholder_dynamic.object = self.dyn_data*self.dyn_raw_data
        else:
            if self.show_widths:
                self._add_integration_windows()
            else:
                self.placeholder_dynamic.object = self.dyn_data*self.dyn_fit_data
    
    @param.depends('corrections',watch = True)
    def _change_correction_button(self):
        if self.corrections:
            #self.get_possible_fine_correction_subshells()
            self.corrections_wid[0].name = 'Count Correcting selected'
            self.corrections_wid[0].button_type = 'success'
            self.edges_correction_width[0].disabled = False
            self.edges_quant_corrector_wid[1].disabled = False
            self._update_possible_corrections()
            
        else:
            self.corrections_wid[0].name = 'No extra-corrections'
            self.corrections_wid[0].button_type = 'danger'
            self._update_possible_corrections()
            self.button_add_corrections.name = '\u2716'
            self.button_add_corrections.button_type = 'danger'
            self.button_add_corrections.disabled = True
            self.edges_correction_width[0].disabled = True
            self.edges_quant_corrector_wid[1].disabled = True
            self.message_display.object = 'No message'
    
    @param.depends('quantification_mode',watch = True)
    def _activate_manual(self):
        #self.show_widths = False
        self._changes_from_quantification_mode()
        self._add_integration_windows()
        
    @param.depends('width_config_mode',watch = True)
    def _changes_for_width_mode(self):
        #self.show_widths = False
        self._changes_from_width_config_mode()
        self._add_integration_windows()
        
    
    @param.depends('elems_quant','edges_for_quant',watch = True)
    def _change_available_sshells(self):
        if len(self.elems_quant) <= 1 or len(self.edges_for_quant) <= 1:
            self.button_quantify.disabled = True
        elif len(self.elems_quant) != len(self.edges_for_quant):
            self.button_quantify.disabled = True
        else:
            self.button_quantify.disabled = False
    
    @param.depends('edges_for_quant',watch = True)
    def _change_edges_available_with_edges_changes(self):
        current_edges_selected = cp.deepcopy(self.edges_for_quant)
        #Whenever we select an specific edge...we want to
        #avoid selecting the other possible edges for the same element
        list_elmental = ['_'.join([ele,name]) for ele in self.elems_quant \
                for name in self.dictio_subshells_names[ele].keys()]
        list_refined = [ed for ed in list_elmental if \
            (all([ed.split('_')[0]+'_' not in clav for clav in current_edges_selected]))]
        for el in current_edges_selected:
            list_refined.append(el)
        self.param['edges_for_quant'].objects = list_refined 
        #We also include the conditional, that when we have at least 2 edges...the
        #manual option is activated and also the rest of configurable tools
        if len(self.edges_for_quant)>=2:
            self.button_show_int_windows[0].disabled = False
            self.quantification_mode_wid[0].disabled = False
            self.param['elems_width'].objects = self.edges_for_quant 
            self.elems_width = self.edges_for_quant[0]
            self._changes_from_quantification_mode()
            self.corrections_wid[0].disabled = False
            self.dictio_possible_corrections = dict()
            try:
                self.get_possible_fine_correction_subshells()
            except: pass
        else:
            self.button_refresh_data.disabled = True
            self.button_reset_data.disabled = True
            self.show_widths = False
            self.button_show_int_windows[0].disabled = True
            self.width_mode_wid[0].disabled = True
            self.onset_int_wid[0].style = {'color':'grey'}
            self.int_width_wid[0].style = {'color':'grey'}
            self.end_int_wid[0].style = {'color':'grey'}
            self.int_width_wid[1].disabled = True
            self.onset_int_wid[1].disabled = True
            self.end_int_wid[1].disabled = True
            self.quantification_mode_wid[0].disabled = True
            self.param['elems_width'].objects = ['EmptyEdgeSelection',]
            self.elems_width = 'EmptyEdgeSelection'
            self.elementos_gos_wid[0].disabled = True
            self.dictionary_for_int = dict()
            self.corrections = False
            #self.edges_correction_width[0].disabled = True
            #self.edges_quant_corrector_wid[1].disabled = True
            self.button_add_corrections.name = '\u2716'
            self.button_add_corrections.disabled = True
            self.corrections_wid[0].disabled = True
            self.dictio_possible_corrections = dict()
    
    def _update_possible_corrections(self):
        #Updates the possible correction list of elements
        if self.corrections:
            #In any case...changes in this button will always reset the corrections dictionary
            self.dictionary_extra_corrections = dict() 
            #We activate the possible options for change
            self.param['elems_extra_corr'].objects = list(self.dictio_possible_corrections.keys())
            self.elems_extra_corr = list(self.dictio_possible_corrections.keys())[0]
            self.param['subshells_for_correction'].objects =\
                ['_'.join([el,edge]) for el in self.dictio_possible_corrections[self.elems_extra_corr] \
                for edge in self.dictio_possible_corrections[self.elems_extra_corr][el]]
        else:
            self.dictionary_extra_corrections = dict()
            self.param['elems_extra_corr'].objects = ['NoneAvailable']
            self.elems_extra_corr = 'NoneAvailable'
            self.subshells_for_correction = []
            self.param['subshells_for_correction'].objects = []
    
    
    
    @param.depends('elems_extra_corr',watch = True)
    def _change_options_ssh(self):
        #When we change between possible correctors
        try:
            lista = self.dictionary_extra_corrections[self.elems_extra_corr]
        except:
            lista = []
        finally:
            self.param['subshells_for_correction'].objects =\
            ['_'.join([el,edge]) for el in self.dictio_possible_corrections[self.elems_extra_corr] \
            for edge in self.dictio_possible_corrections[self.elems_extra_corr][el]]
            self.subshells_for_correction = lista
            
    def _add_corrections(self,event):
        self.button_add_corrections.name = '\u2714'
        self.button_add_corrections.button_type = 'success'
        self.button_add_corrections.disabled = True
        self.dictionary_extra_corrections[self.elems_extra_corr] = self.subshells_for_correction
        lista_message = [' - Corrections in place - <br>']
        for el in self.dictionary_extra_corrections.keys():
            cadena = ' -- Element : {}<br> -- Subshells : {}'.\
            format(el,self.dictionary_extra_corrections[el])
            lista_message.append(cadena)
        self.message_display.object  = ''.join(lista_message)
        
    
    @param.depends('subshells_for_correction',watch = True)
    def _activate_add_corrections_button(self):
        try: 
            long = len(self.subshells_for_correction)
        except:
            long = 0
        finally:
            if long != 0:
                self.button_add_corrections.disabled = False
                #self.button_add_corrections.name = '\u2714' #tick
                #self.button_add_corrections.disabled = True
                self.button_add_corrections.name = 'Add'
                self.button_add_corrections.button_type = 'warning'
            else:
                self.button_add_corrections.button_type = 'warning'
                self.button_add_corrections.name = 'Add'
                #self.button_add_corrections.name = '\u2716'
                #self.button_add_corrections.button_type = 'danger'
                #self.button_add_corrections.disabled = True
    
    def get_possible_fine_correction_subshells(self):
        for edge in self.edges_for_quant:
            ele,shell = edge.split('_')
            corr = False
            self.dictio_possible_corrections[ele] = dict()
            #We want all the other elements, even if they are not selected for quantification
            for el2 in self.dictio_subshells_names.keys():
                #only those elements different than the own element
                if el2 != ele:
                    self.dictio_possible_corrections[ele][el2] = list()
                    for ssh in self.x_sec_funcs[self.xsec][el2]:
                        if ssh not in self.dictio_subshells_names[el2][shell]:
                            #Theses would be the possibilities for corrections of the element ele
                            corr = True
                            self.dictio_possible_corrections[ele][el2].append(ssh)
            if not corr:
                self.dictio_possible_corrections.pop(ele)
            
    @param.depends('elems_width',watch = True)
    def _change_values_onset_and_widths_manual(self):
        if self.elems_width == 'EmptyEdgeSelection':
            #Do nothing and ensure that the widgets is deactivated
            self.elementos_gos_wid[0].disabled = True
            self.int_width_wid[1].disabled = True
            self.onset_int_wid[1].disabled = True
            self.end_int_wid[1].disabled = True
            return
        if self.quantification_mode == 'Auto':
            self.int_width_wid[1].disabled = True
            self.onset_int_wid[1].disabled = True
            self.end_int_wid[1].disabled = True
        else:
            if self.width_config_mode == 'Fixed':
                self.onset_int_wid[1].disabled = False
                self.int_width_wid[1].disabled = False
                self.end_int_wid[1].disabled = True
                ele,sshell = self.elems_width.split('_')
                self.int_window_ev =\
                round(self.dictio_E_lims_manual_h[ele][sshell][1]-self.dictio_E_lims_manual_h[ele][sshell][0],2)
                self.onset_window = self.dictio_E_lims_manual_h[ele][sshell][0]
            elif self.width_config_mode == 'Fluid':
                self.onset_int_wid[1].disabled = False
                self.int_width_wid[1].disabled = True
                self.end_int_wid[1].disabled = False
                ele,sshell = self.elems_width.split('_')
                self.end_window =self.dictio_E_lims_manual_i[ele][sshell][1]
                self.onset_window = self.dictio_E_lims_manual_i[ele][sshell][0]
    
    @param.depends('elems_quant',watch = True)
    def _change_edges_available_with_elemental_changes(self):
        #This solves the problem with the selected elements -
        #the edges will be in the list as long as the elements are selected
        list_elmental = ['_'.join([ele,name]) for ele in self.elems_quant \
                for name in self.dictio_subshells_names[ele].keys()]
        current_selection_options = cp.deepcopy(self.param['edges_for_quant'].objects)
        current_edges_selected = cp.deepcopy(self.edges_for_quant)
        #new_selection_options = [el for el in current_selection_options if el in list_elmental]
        new_edges = [el for el in current_edges_selected if el in list_elmental]
        self.param['edges_for_quant'].objects = list_elmental
        self.edges_for_quant = new_edges
    
    
        
    def create_layout(self):
        self.banner = pn.Row(pn.pane.Markdown('### Quantification Tool',\
            style = {'color':'white'},width = 700,margin = (15,25,0,25)),\
        background='black',width = 1450,height = 70,margin = 0)
        self.button_column =pn.Column(\
            pn.layout.Divider(height = 5,margin = (0,15,5,15)),\
            pn.pane.Markdown('#### Elements,X-section and data for quantification',\
                style = {'color':'white'},margin = (0,15,-10,15),width = 370),\
            self.elementos_quant_wid,self.edges_quant_selector_wid,\
            #self.xsec_sel_wid,
            self.data_sel_wid,\
            pn.layout.Divider(height = 5,margin = (0,15,5,15)),\
            pn.pane.Markdown('#### Extra Electron-counts corrections',\
                style = {'color':'white'},margin = (0,15,-10,15),width = 370),\
            pn.Row(self.corrections_wid,self.button_info_corr,\
                width = 400,margin = 0),\
            self.edges_correction_width,\
            pn.Row(self.edges_quant_corrector_wid,self.button_add_corrections),\
            pn.layout.Divider(height = 5,margin = (0,15,5,15)),\
            self.button_quantify,\
            pn.layout.Divider(height = 5,margin = (-10,15,0,15)),\
            #self.info_panel,\
            height = 655,width =400,margin = 0,background='black')
        self.button_column2 = pn.Column(\
            pn.Row(pn.pane.Markdown('#### Integration width configurator',\
                style = {'color':'white'},margin = (5,15,-10,15),width = 215),\
                self.button_show_int_windows,margin = 0,width = 400),\
            pn.layout.Divider(height = 5,margin = (0,15,5,15)),\
            pn.Row(self.quantification_mode_wid,self.width_mode_wid,\
                width = 400,margin = (5,0)),\
            self.elementos_gos_wid,\
            pn.Row(self.onset_int_wid,self.int_width_wid,\
                width = 400,margin = (5,0)),\
            pn.Row(self.end_int_wid,\
                pn.Row(self.button_refresh_data,self.button_reset_data,margin = 0),\
                width = 400,margin = (5,0)),\
            height = 315,width =400,margin = (0,10),background='black')
        #Placeholders and images
        self.placeholder_dynamic = pn.pane.HoloViews(self.dyn_data*self.dyn_raw_data)
        self.placeholder_image_dyn  = pn.pane.HoloViews(self.im*self.hmap)
        self.init_graphs_panel = pn.Row(self.placeholder_image_dyn,self.placeholder_dynamic,\
            height = 315,width = 1020,margin = (0,5,5,5))
        self.display_raw2 = pn.Row(self.button_column2,self.info_panel,\
            width = 1020,height = 315,margin = (5,5,0,5))
        self.column2 = pn.Column(self.init_graphs_panel,self.display_raw2,\
            height = 620,width = 1000,margin = (15,25,5,25)) 
        self.panel_fila = pn.Row(self.button_column,\
            self.column2,margin = 0,width = 1450)
        
        self.layout = pn.Column(self.banner,self.panel_fila)

    
class Show_results(param.Parameterized):
    base_element = param.ObjectSelector(objects = [])
    use_corrected = param.Boolean(False)
    colormap_ratio = param.ObjectSelector('greys_r',objects = ['greys_r'])
    ratio_selector = param.ObjectSelector('None',objects = ['None'])
    range_clim = param.Range((0,1))
    marker_sel = param.ObjectSelector('circle',\
        objects = ['circle','square','up',\
        'down','diamond','plus'])
    marker_col = param.ObjectSelector()
    show_scat = param.Boolean(default = False)
    
    def __init__(self,ds,ds_counts,dictio_xsec,edges,data_used,int_win_mode,\
        width_mode, extra_corr = False,extra_corr_dict = dict()):
        super().__init__()
        self.ds = xr.merge([ds,ds_counts])
        self.ds.attrs = ds.attrs
        self.dictio_ratios_on_disp = dict()
        self.dictio_shape_markers =\
            {'square':'s','up':'^','down':'v',\
            'diamond':'d','circle':'o','plus':'+'}
        self.dictio_ratio_scatter_plot = dict()
        self.param['marker_col'].objects = bokeh.colors.named.__all__
        xsize = self.ds.x.values.size
        ysize = self.ds.y.values.size
        dist = abs(ysize-xsize)
        if xsize < ysize:
            self.xlims = (-0.5-dist/2,xsize-0.5+dist/2)
            self.ylims = (-0.5,ysize-0.5)
        else:
            self.ylims = (-0.5-dist/2,ysize-0.5+dist/2)
            self.xlims = (-0.5,xsize-0.5)
        #Data to be plotted when clicked
        if data_used == 'Fit - data':
            self.data_clave = 'BestFit'
        else:
            self.data_clave = 'OriginalData'
        ##########################################################
        #The image of the SI original, to inspect spectra later
        TT_coords = [("x","@x"),("y","@y")]
        self.hover_tip_coords = HoverTool(tooltips=TT_coords)
        self.im = hv.Image(self.ds.OriginalData.sum('Eloss'))\
            .opts(cmap = 'greys_r',invert_yaxis = True,aspect = 'equal', frame_height=250,\
            xlim = self.xlims,ylim = self.ylims,\
            xaxis = None,yaxis = None,alpha = 0,shared_axes = False,toolbar = 'left')
        self.hmap = hv.HeatMap(self.ds.OriginalData.sum('Eloss'))\
            .opts(cmap = 'greys_r',invert_yaxis = True,aspect = 'equal',frame_height=250,\
            xlim = self.xlims,ylim = self.ylims,\
            bgcolor='black',line_width = 1,line_alpha = 0.15,\
            selection_line_alpha = 1,\
            xaxis = None,yaxis = None,selection_fill_alpha = 1,\
            nonselection_fill_alpha = 0.75,\
            line_color = 'grey',selection_line_color = 'red',\
            nonselection_line_color = 'grey',alpha = 0.75,\
            hover_fill_alpha = 1, hover_line_color = 'limegreen',\
            tools = [self.hover_tip_coords],shared_axes = False,toolbar = 'left')
        self.hov_lims = (self.ds.x.values[0]-0.5,self.ds.x.values[-1]+\
                0.5,self.ds.y.values[0]-0.5,self.ds.y.values[-1]+0.5)
        #self.stream_tapping  = streams.SingleTap(x = -1,y = -1,source = self.hmap)
        #self.stream_hovering = streams.PointerXY(x = -1,y = -1,source = self.im)
        #self.dyn_tap = hv.DynamicMap(self._plot_area_data,streams=[self.stream_tapping])\
        #   .opts(framewise = True,shared_axes = False,frame_height=250,frame_width =500)
        ###########################################################
        #Information displayed
        self.mkdw_data_used =\
            pn.pane.Markdown('#### Data used for integration : {}'.format(data_used),\
            style = {'color':'white'},margin = (0,25),width = 350)
        self.mkdw_int_win_mode =\
            pn.pane.Markdown('#### Integration Mode : {}'.format(int_win_mode),\
            style = {'color':'white'},margin = (0,25),width = 350)
        self.mkdn_window_width_selection =\
            pn.pane.Markdown('#### Integration Mode : {}'.format(width_mode),\
            style = {'color':'white'},margin = (0,25),width = 350)
        self.button_update_cmap = pn.widgets.Button(name = '\u27F3',disabled = True,\
            button_type = 'warning',height = 33,width = 30,margin = (33,10,5,10))
        self.button_update_cmap.on_click(self._callback_update_cmap)
        self.button_update_scatter = pn.widgets.Button(name = '\u2317',disabled = True,\
            button_type = 'primary',height = 33,width = 30,margin = (33,10,5,10))
        self.button_update_scatter.on_click(self._callback_update_scatter)
        self.button_use_corrected_counts = pn.Param(self.param['use_corrected'],\
            widgets = {'use_corrected':pn.widgets.Toggle},parameters = ['use_corrected'],\
            show_name = False,show_labels = False,width = 310,align = 'center',margin = 0)
        self.button_use_corrected_counts[0].width = 300
        self.button_use_corrected_counts[0].margin = 5
        self.button_use_corrected_counts[0].name = 'No corrections'
        self.button_use_corrected_counts[0].button_type = 'danger'
        self.button_save_results = pn.widgets.Button(name = 'Save Data-Images',\
            button_type = 'danger',disabled = True,width = 250,height = 35,margin = (5,50),\
            css_classes = ['custom_button_bokeh_black_inter',\
            'custom_button_bokeh_black_inter_dis'])
        self.button_save_results.on_click(self._callback_saving_window)
        self.dictio_corrections = cp.deepcopy(extra_corr_dict)
        if extra_corr and extra_corr_dict != dict():
            extra_corrections = []
            for i,el in enumerate(extra_corr_dict):
                lista = [' '.join(ele.split('_')) for ele in extra_corr_dict[el]]
                extra_corrections.append(
                ' - Element ({}) - {} -|- Cross-sections : {}'.format(i+1,el,','.join(lista)))
            self.mkdw_extra_corrections =\
                pn.pane.Markdown('#### Extra Corrections in place:<br> {}'.\
                format('<br>'.join(extra_corrections)),width = 350,\
                style = {'color':'white'},margin = (0,25))
            self.button_use_corrected_counts[0].disabled = False
        else:
            self.mkdw_extra_corrections =\
            pn.pane.Markdown('#### Extra Corrections in place:<br> {}'.\
                format('- None -'),width = 350,\
                style = {'color':'white'},margin = (0,25))
            self.button_use_corrected_counts[0].disabled = True
        #NOw let's get the values for the possible base elements
        self.dictio_xsec = dictio_xsec
        self.param['base_element'].objects = list(dictio_xsec.keys())
        #self.base_element = self.param['base_element'].objects[0]
        self.base_element_wid = pn.Param(self.param['base_element'],\
            widgets = {'base_element':pn.widgets.RadioButtonGroup},\
            parameters = ['base_element'],align = 'center',\
            show_name = False,show_labels = False,width = 300)
        self.base_element_wid[0].margin = 0
        self.base_element_wid[0].width = 300
        self.base_element_wid[0].button_type = 'success'
        self.button_compute = pn.widgets.Button(name = 'Show ratios',\
            width = 300,align = 'center',disabled = False)
        self.button_compute.on_click(self._show_ratios)
        '''
        self.button_show_scatter = pn.widgets.Button(name = 'Show line ratios',\
            width = 150,disabled = True,margin = (5,10),button_type = 'primary')
        self.button_show_scatter.on_click(self._callback_show_scat)
        '''
        #Now the always useful empty data image
        empty_Data =\
            xr.Dataset({'EmptyData':(['y','x','Eloss'],\
            np.zeros_like(self.ds.OriginalData.values))},\
            coords = {'y':self.ds.y.values,'x':self.ds.x.values,\
            'Eloss':self.ds.Eloss.values})
        self.curve_empty = hv.Curve(empty_Data.EmptyData.isel(x = 0 ,y = 0))\
        .opts(frame_height = 250,frame_width = 500,show_grid = True,\
            shared_axes = False,yformatter=formatter,framewise = True,\
            line_width = 0.1,line_color = 'black')
        self.area_empty = hv.Area(empty_Data.EmptyData.isel(x = 0 ,y = 0))\
        .opts(frame_height = 250,frame_width = 500,show_grid = True,\
            shared_axes = False,yformatter=formatter,\
            line_width = 0.1,fill_alpha = 0.1,fill_color = 'grey',\
            line_color = 'black',framewise = True)
        self.empty_hmap = hv.HeatMap(empty_Data.sum('Eloss'))\
            .opts(cmap = 'greys_r',invert_yaxis = True,aspect = 'equal',\
            frame_height=250,xlim = self.xlims,ylim = self.ylims,\
            xaxis = None,yaxis = None,selection_fill_alpha = 1,\
            nonselection_fill_alpha = 0.75,line_width = 1.25,\
            line_color = 'grey',selection_line_color = 'red',\
            nonselection_line_color = 'grey',alpha = 0.75,\
            hover_fill_alpha = 1, hover_line_color = 'limegreen',\
            tools = [self.hover_tip_coords],shared_axes = False,\
            colorbar = True,colorbar_position = 'right',toolbar = 'left',\
            show_title = True,title = 'empty')
        self.scatter_empty = hv.Scatter(\
            xr.Dataset({'Empty_scatter':(['x'],np.ones(10))},\
            coords = {'x':np.arange(0,10,1)}))\
            .opts(frame_height = 250,frame_width = 500,show_grid = True,yformatter=formatter4,\
                shared_axes = False,framewise = True)
        self.path = hv.Path([[(0,0),(0,0)]]).opts(color = 'gold',\
            line_width = 2.5,line_alpha = 0.5)
        self.path_stream =\
        streams.PolyDraw(source=self.path, drag=True,\
            vertex_style = {'color':'gold','line_color':'orangered'},\
            show_vertices=True,num_objects=1)
        '''
        self.dyn_draw = hv.DynamicMap(self._plot_ratio_line,streams=[self.path_stream])\
            .opts(frame_height = 250,shared_axes = False,frame_width = 500,\
            framewise = True,\
            yformatter=formatter4)
        self.mkdw_rat_on_display = pn.pane.Markdown('#### Ratios on display:<br>None',\
        style = {'color':'lightgrey'},margin = (0,15))
        '''
        #Some widgets_modifications
        self.cmap_wid = pn.Param(self.param['colormap_ratio'],\
            widgets = {'colormap_ratio':pn.widgets.Select},\
            parameters = ['colormap_ratio'],name = 'Colormap',show_name = True,\
            show_labels = False,width = 90, margin = 5)
        self.cmap_wid[0].style = {'color':'white'}
        self.cmap_wid[1].margin = 0
        self.cmap_wid[0].margin = 5
        self.cmap_wid[1].width = 90
        self.cmap_wid[0].width = 80
        self.cmap_wid[1].disabled = True
        self.ratio_selector_wid = pn.Param(self.param['ratio_selector'],\
            widgets = {'ratio_selector':pn.widgets.Select},\
            parameters = ['ratio_selector'],show_name = False,\
            show_labels = False,width = 125,margin = (5,10))
        #self.ratio_selector_wid[0].style = {'color':'white'}
        self.ratio_selector_wid[0].disabled = True
        self.range_clim_wid = pn.Param(self.param['range_clim'],\
            widgets = {'range_clim':pn.widgets.RangeSlider},\
            parameters = ['range_clim'],width = 90,margin = 5,\
            show_labels = False,show_name = True,\
            name = 'Limits')
        self.range_clim_wid[0].style = {'color':'white'}
        self.range_clim_wid[1].bar_color = '#FF7F50'
        self.range_clim_wid[1].step = 0.01
        self.range_clim_wid[1].show_value = False
        self.range_clim_wid[0].width = 80
        self.range_clim_wid[1].width = 90
        self.range_clim_wid[0].margin = 5
        self.range_clim_wid[1].margin = 0
        self.range_clim_wid[1].disabled = True

        #Marker options for scatter plots
        self.marker_col_wid = pn.Param(self.param['marker_col'],\
            widgets = {'marker_col':pn.widgets.Select},parameters = ['marker_col'],\
            show_name = True,name = 'Color',\
            show_labels = False,width = 90,margin = 5)
        self.marker_sel_wid = pn.Param(self.param['marker_sel'],\
            widgets = {'marker_sel':pn.widgets.Select},parameters = ['marker_sel'],\
            show_name = True,name = 'Marker',\
            show_labels = False,width = 90,margin = 5)
        self.marker_sel_wid[0].style = {'color':'white'}
        self.marker_col_wid[0].style = {'color':'white'}
        self.marker_sel_wid[1].margin = 0
        self.marker_col_wid[1].margin = 0
        self.marker_sel_wid[0].margin = 5
        self.marker_col_wid[0].margin = 5
        self.marker_sel_wid[0].width = 80
        self.marker_col_wid[0].width = 80
        self.marker_sel_wid[1].width = 90
        self.marker_col_wid[1].width = 90
        self.marker_col_wid[1].disabled = True
        self.marker_sel_wid[1].disabled = True
        self.scatter_inplace = False
    
    

    def algorithm_DDA(self,xi,xf,yi,yf):
        #Digital differential analyzer (DDA) - getting the pixels in the line
        dx = (xf-xi)
        dy = (yf-yi)
        if abs(dx) >= abs(dy):
            step = abs(dx)
        else:
            step = abs(dy)
        dx /= step
        dy /= step
        list_x = []
        list_y = []
        idx_im = []
        i = 1
        x = xi
        y = yi
        while i<= step:
            list_x.append(x)
            list_y.append(y)
            x += dx
            y += dy
            #to make them integers
            #x,y = im.closest((x,y))
            i += 1
        #At the end, we correct to get the actual pixel possitions
        for i,el in enumerate(list_x):
            #We are applying this to the image we are interest on
            idx_im.append(self.im.closest((el,list_y[i])))
        x,y = self.im.closest((xf,yf))
        idx_im.append((x,y))
        return idx_im
    

    def _plot_ratio_line(self,data,refresh = False):
        #try:
        if len(data['xs']) != 1:
            #self.show_scat = False
            print('case1')
            '''
            nover = hv.NdOverlay({'No-data':self.scatter_empty})\
            .opts(shared_axes = False,framewise = True,\
                xlabel = 'Pixel position along the line',ylabel = 'Elemental ratio') 
            self.dyn_placement.object = nover
            '''
            return self.scatter_empty
        elif data['xs'][0][0] == data['xs'][0][1] and data['ys'][0][0] == data['ys'][0][1]:
            #print('case2')
            self.show_scat = False
            #Case of not having an entire segment yet - selection pending of completion
            '''
            nover = hv.NdOverlay({'No-data':self.scatter_empty})\
            .opts(shared_axes = False,framewise = True,\
                xlabel = 'Pixel position along the line',ylabel = 'Elemental ratio') 
            self.dyn_placement.object = nover
            '''
            return self.scatter_empty
        else:
            #print('case3')
            self.show_scat = True
            '''
            list_segment_coords = self.algorithm_DDA(*data['xs'][0],*data['ys'][0])
            scat_dict = dict()
            for i,gr in enumerate(self.grid_images):
                dat = list(gr.object)[0].data
                vals = []
                for pon in list_segment_coords:
                    vals.append(round(dat['z'].values[int(pon[1]),int(pon[0])],2))
                if refresh and self.ratio_selector == self.param['ratio_selector'].objects[i]:
                    colour, mark = [self.marker_col,self.marker_sel]
                else:
                    colour, mark = self.dictio_ratio_scatter_plot[self.param['ratio_selector'].objects[i]]
                scat_dict[self.param['ratio_selector'].objects[i]] = \
                    hv.Scatter(xr.Dataset({self.param['ratio_selector'].objects[i]:\
                    ('x',np.array(vals))},coords={'x':np.arange(0,len(vals))}))\
                    .opts(size = 10,frame_height = 250,frame_width = 500,show_grid = True,\
                        line_color = colour,line_alpha = 1, fill_alpha = 0.5,fill_color = colour,\
                        shared_axes = False,yformatter=formatter4,\
                        framewise = True,\
                        marker = self.dictio_shape_markers[mark])
                if not refresh:
                    new_path = hv.Path(self.algorithm_DDA(*data['xs'][0],*data['ys'][0]))\
                    .opts(line_width = 3,line_color = colour)
                    gr.object = list(gr.object)[0]*new_path
                else: pass
            scat = hv.NdOverlay(scat_dict).opts(shared_axes = False,framewise = True,\
                xlabel = 'Pixel position along the line',ylabel = 'Elemental ratio')
            self.dyn_placement.object = scat
            if not self.scatter_inplace:
                self.scatter_inplace = True
                self.marker_col_wid[1].disabled = False
                self.marker_sel_wid[1].disabled = False
                self.button_update_scatter.disabled = False
            else:pass
            '''
            return self.scatter_empty
    '''
    @param.depends('show_scat',watch = True)
    def _allow_scat(self):
        #if self.show_scat:
        self.button_show_scatter.disabled = not self.show_scat
    
    def _callback_show_scat(self,event):
        self._plot_scatter_refresh(self.path_stream.data)
        self.show_scat = False
    '''     
    
    def _plot_area_data(self,x,y):
        xi,xf,yi,yf = self.hov_lims
        if x >= xf or x <= xi:
            return self.area_empty #*self.curve_empty
        elif y >= yf or y <= yi:
            return self.area_empty #*self.curve_empty
        else:
            x0,y0 = self.im.closest((x,y))
            self.x_pos = int(x0)
            self.y_pos = int(y0)
            area = hv.Area(self.ds[self.data_clave].isel(x = int(x0),y = int(y0)))\
                .opts(show_grid=True,show_title=False,line_width=0.1,\
                yformatter=formatter,line_alpha = 0,\
                shared_axes = False,framewise = True,fill_color = 'coral',\
                fill_alpha = 1)
            '''
            curva = hv.Curve(self.ds[].isel(x = int(x0),y = int(y0)),\
                label ='Fitted-Data')\
                .opts(show_grid=True,show_title=False,line_width=1,\
                line_color = 'darkred',yformatter=formatter,
                shared_axes = False,framewise = True)
            '''
            return area

    def _show_ratios(self,event):
        #First thing is first - we have to get the actual ratios
        #Lists of sequantial colormaps, well balanced
        #print('yey1')
        #First ... if no element was selected as base...let's pick one of them
        if not self.base_element:
            self.base_element = self.param['base_element'].objects[0]
        else: pass
        lista_cmap_hv = hv.plotting.util.list_cmaps(category='Uniform Sequential')
        lista_cmap =\
        list(set([cm.casefold() for cm in lista_cmap_hv if '_r' not in cm]))
        self.ratio_selector_wid[0].disabled = False
        self.cmap_wid[1].disabled = False
        self.param['colormap_ratio'].objects = cp.deepcopy(lista_cmap)
        #print('yey2')
        self.lst = [el for el in self.param['base_element'].objects \
            if el != self.base_element]
        self.param['ratio_selector'].objects =\
            [' {} / {} '.format(el,self.base_element) for el in self.lst]
        #listo = ['/'.join([el,self.base_element]) for el in lst]
        '''
        self.mkdw_rat_on_display.object =\
        '### Ratios on display:<br>{}'.format(' , '.join(listo))
        '''
        if self.use_corrected and self.base_element in self.dictio_corrections.keys():
            clav_base = '_'.join([self.base_element,'_corrected_intensity'])
        else:
            clav_base = '_'.join([self.base_element,'_integrated_intensity'])
        for i,el in enumerate(self.lst):
            #Let's check we didnt emptied the whole colormap set
            '''
            if len(lista_cmap) == 0:
                #Reset
                lista_cmap =\
                list(set([el.casefold() for el in lista_cmap_hv if '_r' not in el]))
            '''
            #Let's create the image - the dataset 
            if self.use_corrected and el in self.dictio_corrections.keys():
                clav = '_'.join([el,'_corrected_intensity'])
            else:
                clav = '_'.join([el,'_integrated_intensity'])
            #Let's create a mask to avoid further problems
            mascara = np.zeros_like(self.ds.mask.values)
            mascara[self.ds.mask.values!= 0] = 1
            datArr = ((self.ds[clav]*mascara * self.dictio_xsec[self.base_element]['val'])/\
            (self.ds[clav_base]*mascara * self.dictio_xsec[el]['val']))
            titulo ='{}/{} - Relative composition '.format(el,self.base_element)
            cmap_using = choice(lista_cmap)
            #lista_cmap.pop(lista_cmap.index(cmap_using))
            maxi = min(np.nanmax(datArr),10) # If any infinite slips through, we can avoid problems here
            mini = max(np.nanmin(datArr),0)
            #self.range_clim = (max(mini,0),maxi)
            im = hv.HeatMap(datArr)\
            .opts(cmap = cmap_using,\
            invert_yaxis = True,aspect = 'equal',\
            frame_height=250,xlim = self.xlims,ylim = self.ylims,\
            xaxis = None,yaxis = None,selection_fill_alpha = 1,\
            nonselection_fill_alpha = 1,\
            clipping_colors = {'NaN': 'black'},line_width = 0.05,bgcolor = 'darkgrey',\
            line_color = 'white',selection_line_color = 'red',\
            nonselection_line_color = 'white',alpha = 0.75,\
            hover_fill_alpha = 1, hover_line_color = 'limegreen',\
            tools = [self.hover_tip_coords],shared_axes = False,\
            colorbar = True,colorbar_position = 'right',toolbar = 'left',\
            show_title = True,title = titulo,clim = (max(mini,0),maxi))
            self.grid_images[i].object = im*hv.Path([])
            self.dictio_ratios_on_disp[self.param['ratio_selector'].objects[i]] =\
                [i,cmap_using,(round(max(mini,0),2),round(maxi,2))]
            #Options for the latter scatter plots
            colour = choice(bokeh.colors.named.__all__)
            marker_key = choice(self.param['marker_sel'].objects)
            self.dictio_ratio_scatter_plot[self.param['ratio_selector'].objects[i]] =\
                [colour,marker_key]
            #self.ratio_selector = self.param['ratio_selector'].objects[0]
        self.button_update_cmap.disabled = False
        self.range_clim_wid[1].disabled = False
        self.ratio_selector = self.param['ratio_selector'].objects[0]
        self.button_update_scatter.disabled = False
        self.button_save_results.disabled = False
        self.button_save_results.button_type = 'default'
        #self.range_clim = self.dictio_ratios_on_disp[self.ratio_selector][2]

    def _plot_scatter_refresh(self,data,refresh = False):
        if len(data['xs']) != 1:
            return True
        elif data['xs'][0][0] == data['xs'][0][1] and data['ys'][0][0] == data['ys'][0][1]:
            return True
        else:
            list_segment_coords = self.algorithm_DDA(*data['xs'][0],*data['ys'][0])
            scat_dict = dict()
            for i,gr in enumerate(self.grid_images):
                dat = list(gr.object)[0].data
                vals = []
                for pon in list_segment_coords:
                    vals.append(round(dat['z'].values[int(pon[1]),int(pon[0])],2))
                if refresh and self.ratio_selector == self.param['ratio_selector'].objects[i]:
                    colour, mark = [self.marker_col,self.marker_sel]
                    self.dictio_ratio_scatter_plot[self.param['ratio_selector'].objects[i]] =\
                        [colour,mark]
                else:
                    colour, mark = self.dictio_ratio_scatter_plot[self.param['ratio_selector'].objects[i]]
                scat_dict[self.param['ratio_selector'].objects[i]] =\
                    hv.Scatter(\
                    xr.Dataset({self.param['ratio_selector'].objects[i]:\
                    ('x',np.array(vals))},coords={'x':np.arange(0,len(vals))}))\
                    .opts(size = 10,frame_height = 250,frame_width = 500,show_grid = True,\
                        line_color = colour,line_alpha = 1, fill_alpha = 0.5,fill_color = colour,\
                        shared_axes = False,yformatter=formatter4,\
                        framewise = True,marker = self.dictio_shape_markers[mark])
                new_path = hv.Path(self.algorithm_DDA(*data['xs'][0],*data['ys'][0]))\
                    .opts(line_width = 3,line_color = colour)
                gr.object = list(gr.object)[0]*new_path
            scat = hv.NdOverlay(scat_dict).opts(shared_axes = False,framewise = True,\
                xlabel = 'Pixel position along the line',ylabel = 'Elemental ratio')
            self.dyn_placement.object = scat
            return False

    def _callback_saving_window(self,event):
        #Let's get all the graphs needed here
        #Let's take the images and names

        lista_images = [hv.Overlay.clone(el.object) for el in self.grid_images]
        lista_nombres = cp.deepcopy(self.lst)
        lista_images.append(hv.NdOverlay.clone(self.dyn_placement.object))
        lista_nombres.append('LineScan')
        sv = Saving_panel(ds = self.ds,\
            figures = lista_images, figures_names = lista_nombres,\
            name_panel = 'quantification')
        sv.create_layout()
        

    def _callback_update_scatter(self,event):
        '''
        num = self.dictio_ratios_on_disp[self.ratio_selector][0]
        imi = list(self.grid_images[num].object)[0]
        path = list(self.grid_images[num].object)[1]
        self.grid_images[num].object = imi * path\
        .opts(line_color = self.marker_col)
        '''
        out = self._plot_scatter_refresh(self.path_stream.data,refresh = self.show_scat)
        if out:
            return
        else: pass
        if not self.show_scat:
            self.show_scat = True
            self.marker_col_wid[1].disabled = False
            self.marker_sel_wid[1].disabled = False
            self.button_update_scatter.name = '\u27F3'
            self.button_update_scatter.button_type = 'danger'
        else: 
            pass
        

    def _callback_update_cmap(self,event):
        num = self.dictio_ratios_on_disp[self.ratio_selector][0]
        imi = list(self.grid_images[num].object)[0]
        path = list(self.grid_images[num].object)[1]
        self.grid_images[num].object =\
        imi.opts(cmap = self.colormap_ratio,clim= self.range_clim) * path
        self.dictio_ratios_on_disp[self.ratio_selector] =\
                [num,self.colormap_ratio,self.range_clim]
        
    @param.depends('ratio_selector',watch = True)
    def change_colormap_options(self):
        if self.ratio_selector != 'None':
            #For the heatmaps
            rat_pars = self.dictio_ratios_on_disp[self.ratio_selector]
            mini,maxi = rat_pars[2]
            dif = maxi-mini
            self.param['range_clim'].bounds =\
                (round(max(mini-0.5*dif,0),2),round(maxi+0.5*dif,2))
            self.colormap_ratio = rat_pars[1]
            self.range_clim = self.dictio_ratios_on_disp[self.ratio_selector][2]
            #For the scatter plots
            rat_scatt = self.dictio_ratio_scatter_plot[self.ratio_selector] 
            self.marker_col = rat_scatt[0]
            self.marker_sel = rat_scatt[1]
        else:
            return

    @param.depends('use_corrected',watch = True)
    def change_corr_button_style(self):
        if self.use_corrected:
            self.button_use_corrected_counts[0].name = 'Using corrections'
            self.button_use_corrected_counts[0].button_type = 'success'
        else:
            self.button_use_corrected_counts[0].name = 'No corrections'
            self.button_use_corrected_counts[0].button_type = 'danger'

    def create_display(self):
        nover = hv.NdOverlay({'No-data':self.scatter_empty})\
            .opts(shared_axes = False,framewise = True,\
            xlabel = 'Pixel position along the line',ylabel = 'Elemental ratio') 
        self.dyn_placement = pn.pane.HoloViews(nover)
        self.im_placement = pn.pane.HoloViews(self.hmap*self.path)
        self.inspect_graph_line = pn.Column(\
            pn.pane.Markdown('### Dataset quantified',margin = (0,25,-10,25)),\
            pn.Row(self.im_placement,self.dyn_placement),margin = (0,15))
        lista_positional_hmaps = []
        for _ in range(len(self.dictio_xsec.keys())-1):
            lista_positional_hmaps.append(self.empty_hmap)
        self.grid_images = pn.GridBox(*lista_positional_hmaps,nrows = 1,\
            ncols = None,margin = (5,15))
        self.column_info = pn.Column(\
            pn.pane.Markdown(\
                '### Information - Dataset used for quantification',\
                width = 370,style = {'color':'white'},margin = (0,15)),\
            self.mkdw_data_used,self.mkdw_int_win_mode,self.mkdn_window_width_selection,\
            self.mkdw_extra_corrections,\
            pn.layout.Divider(height = 5,margin = (5,10,5,10)),\
            pn.pane.Markdown('#### Base element selector<br> and extra components',\
                width = 300,height = 25,style = {'color':'white'},margin = (0,50)),\
            self.base_element_wid,
            self.button_use_corrected_counts,\
            pn.layout.Divider(height = 5,margin = (5,10,5,10)),\
            self.button_compute,\
            pn.layout.Divider(height = 5,margin = (5,10,5,10)),\
            pn.Row(pn.pane.Markdown('#### Ratio ',width = 35,\
                    style = {'color':'white'},margin = (5,0)),\
                self.ratio_selector_wid,\
                pn.pane.Markdown('#### plotting options ',margin = 5,\
                    width = 125,style = {'color':'white'},background = None),\
                width = 325,margin = (0,25,0,25)),\
            pn.Row(pn.pane.Markdown('#### **HeatMaps**',\
                    style = {'color':'white'},margin = (35,5,0,0),width = 75),\
                self.button_update_cmap,self.cmap_wid,self.range_clim_wid,\
                width = 350,margin = (0,25)),\
            pn.Row(pn.pane.Markdown('#### **Scatter**',\
                    style = {'color':'white'},margin = (35,5,0,0),width = 75),\
                self.button_update_scatter,self.marker_sel_wid,\
                self.marker_col_wid,width = 350,margin = (0,25)),\
            width = 400,margin = (0,5),height = 680,background = 'black')
        self.col_images = pn.Column(self.inspect_graph_line,self.grid_images)
        self.banner = pn.Row(pn.Spacer(height = 45,width = 1150,\
            margin = 0),self.button_save_results,\
            width = 1500,margin = (0,5),background = 'black')
        self.display = pn.Column(\
            self.banner,pn.Row(self.column_info,self.col_images))